#include "Exp.h"
#include <string>
#include <sstream> 

#if!defined( __EXP_USED_BY_COMPILER)
//#include <MEMORY>

bool CExp::readINI(const char *FileName, const long nExpId )
{
	FILE *fp = fopen( FileName, "rb" );
	
	if ( fp == NULL )
	{
		fprintf( stderr, "ExpINIļʧ...\n");
		return false;
	}
	
	char szTemp[MAX_NUM_LEN+sizeof("exp-")] = "exp-";

	ltoa( nExpId, szTemp+strlen(szTemp), 10 );

	if( !OrientationReadPos( fp, "Expressions", szTemp ) )
	{
		fclose( fp );
		return false;
	}
	
	fread( &m_nVarNum, sizeof(m_nVarNum), 1, fp );
	fread( &m_nTreeSize, sizeof(m_nTreeSize), 1, fp );
	fread( &m_nRoot, sizeof(m_nRoot), 1, fp );

	if ( m_nCurrTreeSize < m_nTreeSize )
	{
		if ( m_pBinaryTree!=NULL )
		{
			delete [] m_pBinaryTree;
		}

		m_pBinaryTree = new CNode[m_nCurrTreeSize = m_nTreeSize];
	}

	if ( m_nCurrVarNum < m_nVarNum )
	{
		if ( m_pszVar != NULL )
		{
			delete [] m_pszVar;
		}

		m_pszVar = new VAR[m_nCurrVarNum = m_nVarNum];
	}

	fread( m_pszVar, sizeof(VAR), m_nVarNum, fp );
	fread( m_pBinaryTree, sizeof(CNode), m_nTreeSize, fp );
	
	fclose( fp );

	JudgeExpType();

	SetVarId2BinaryTree();

	return true;
}

bool CExp::readINI(const char *FileName, const char *lpAppName, const char *lpKeyName )
{
	FILE *fp = fopen( FileName, "rb" );
	
	if ( fp == NULL )
	{
		fprintf( stderr, "ExpINIļʧ...\n");
		return false;
	}
	
	if( !OrientationReadPos( fp, lpAppName, lpKeyName ) )
	{
		fclose( fp );
		return false;
	}

	long nExpId = 0;

	fscanf( fp, "%d", &nExpId );

	char szTemp[MAX_NUM_LEN+sizeof("exp-")] = "exp-";

	ltoa( nExpId, szTemp+strlen(szTemp), 10 );

	if( !OrientationReadPos( fp, "Expressions", szTemp ) )
	{
		fclose( fp );
		return false;
	}
	
	fread( &m_nVarNum, sizeof(m_nVarNum), 1, fp );
	fread( &m_nTreeSize, sizeof(m_nTreeSize), 1, fp );
	fread( &m_nRoot, sizeof(m_nRoot), 1, fp );

	if ( m_nCurrTreeSize < m_nTreeSize )
	{
		if ( m_pBinaryTree!=NULL )
		{
			delete [] m_pBinaryTree;
		}

		m_pBinaryTree = new CNode[m_nCurrTreeSize = m_nTreeSize];
	}

	if ( m_nCurrVarNum < m_nVarNum )
	{
		if ( m_pszVar != NULL )
		{
			delete [] m_pszVar;
		}

		m_pszVar = new VAR[m_nCurrVarNum = m_nVarNum];
	}

	fread( m_pszVar, sizeof(VAR), m_nVarNum, fp );
	fread( m_pBinaryTree, sizeof(CNode), m_nTreeSize, fp );
	
	fclose( fp );

	JudgeExpType();

	SetVarId2BinaryTree();

	return true;
}

void CExp::SetVarId2BinaryTree()
{
	for(long i=0; i< m_nVarNum; i++ )
	{
		m_pBinaryTree[m_pszVar[i].m_nPos2Set].m_nId = m_pszVar[i].m_nVarId;
	}
}

ctaState *CExp::m_spCurrState = NULL;

void CExp::SetCurrentState(const ctaState *pState )
{
	m_spCurrState = (ctaState*)pState;
}

void CExp::JudgeExpType()
{
	if ( m_nVarNum == 0 )
	{
		m_ExpOpNumType = ConstExp;
	}
	else
	{
		switch( m_nTreeSize )
		{
		case 1:
			m_ExpOpNumType = OneVarExp;
			break;
		case 2:
			m_ExpOpNumType = OneOpExp;
			break;
		case 3:
			m_ExpOpNumType = TwoOpExp;
			break;
		default:
			m_ExpOpNumType = OverTowOpExp;
			break;
		}
	}

	switch( m_ExpOpNumType )
	{
	case ConstExp:
		{
			const static class CConstExp:public CExp
			{
				long Compute()
				{
					return m_pBinaryTree[0].m_nVal;//return printf( "ConstExp Compute\n" );
				}
			} ConstExpObj;

			memcpy( this, &ConstExpObj, sizeof(void*) );
		}
		break;
	case OneVarExp:
		{
			const static class COneVarExp:public CExp
			{
				long Compute()
				{
					return m_pBinaryTree[0].m_nVal;//return printf( "ConstExp Compute\n" );
				}
			} OneVarExpObj;
			
			memcpy( this, &OneVarExpObj, sizeof(void*) );
		}
		break;
	case OneOpExp:
		{
			const static class COneOpExp:public CExp
			{
				long Compute()
				{
					if( m_pBinaryTree[m_nRoot].m_OpType == '!')
					{
						return !m_pBinaryTree[m_nRoot].m_nVal;
					}
					else
					{
						return -m_pBinaryTree[m_nRoot].m_nVal;
					}
					//return printf( "OneOpExp Compute\n" );
				}
			} OneOpExpObj;

			memcpy( this, &OneOpExpObj, sizeof(void*) );
		}
		break;
	case TwoOpExp:
		{
			static const class CTwoOpExp:public CExp
			{
				long Compute()
				{
					switch( m_pBinaryTree[m_nRoot].m_OpType )
					{
					case '+':
						return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nLeft].m_nVal
							+ m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight].m_nVal;
						break;
					case '-':
						return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nLeft].m_nVal
							- m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight].m_nVal;
						break;
					case '*':
						return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nLeft].m_nVal
							* m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight].m_nVal;
						break;
					case '/':
						return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nLeft].m_nVal
							/ m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight].m_nVal;
						break;
					case '%':
						return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nLeft].m_nVal
							% m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight].m_nVal;
						break;
					case '&':
						return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nLeft].m_nVal
							&& m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight].m_nVal;
						break;
					case '|':
						return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nLeft].m_nVal
							|| m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight].m_nVal;
						break;
					case opGT:
						return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nLeft].m_nVal
							> m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight].m_nVal;
						break;
					case opGE:
						return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nLeft].m_nVal
							>= m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight].m_nVal;
						break;
					case opLT:
						return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nLeft].m_nVal
							< m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight].m_nVal;
						break;
					case opLE:
						return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nLeft].m_nVal
							<= m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight].m_nVal;
						break;
					case opEQ:
						return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nLeft].m_nVal
							== m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight].m_nVal;
						break;
					case opNotEQ:
						return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nLeft].m_nVal
							!= m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight].m_nVal;
						break;
					default:
						return 0;
						break;
					}
					//return printf( "TwoOpExp Compute\n" );
				}

				long UpdateAndCmptRightVal()
				{
					UpdateVal();
					return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight].m_nVal;
				}

			} TwoOpExpObj;

			memcpy( this, &TwoOpExpObj, sizeof(void*) );

			//adds 2007/02/09
			if ( m_nVarNum == 1 && m_pszVar[0].m_nPos2Set != m_pBinaryTree[m_nRoot].m_nRight )//ұǳ
			{
				static const class CTwoOpExpEx:public CTwoOpExp
				{
					long UpdateAndCmptRightVal()
					{
						return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight].m_nVal;
					}
				}TwoOpExpObjEx;

				memcpy( this, &TwoOpExpObjEx, sizeof(void*) );
			}
		}
		break;
	case OverTowOpExp:
		{
			static const class COverTwoOpExp:public CExp
			{
				long Compute()
				{
					return Compute( m_pBinaryTree[m_nRoot] );
					//return printf( "OverTwoOpExp Compute\n" );
				}

				long Compute(CNode &Root)
				{
					switch( Root.m_OpType )
					{
					case '+':
						return Compute(m_pBinaryTree[Root.m_nLeft])
							+ Compute(m_pBinaryTree[Root.m_nRight]);
						break;
					case '-':
						return Compute(m_pBinaryTree[Root.m_nLeft])
							- Compute(m_pBinaryTree[Root.m_nRight]);
						break;
					case '*':
						return Compute(m_pBinaryTree[Root.m_nLeft])
							* Compute(m_pBinaryTree[Root.m_nRight]);
						break;
					case '/':
						return Compute(m_pBinaryTree[Root.m_nLeft])
							/ Compute(m_pBinaryTree[Root.m_nRight]);
						break;
					case '%':
						return Compute(m_pBinaryTree[Root.m_nLeft])
							% Compute(m_pBinaryTree[Root.m_nRight]);
						break;
					case '&':
						return Compute(m_pBinaryTree[Root.m_nLeft])
							&& Compute(m_pBinaryTree[Root.m_nRight]);
						break;
					case '|':
						return Compute(m_pBinaryTree[Root.m_nLeft])
							|| Compute(m_pBinaryTree[Root.m_nRight]);
						break;
					case opLT:
						return Compute(m_pBinaryTree[Root.m_nLeft])
							< Compute(m_pBinaryTree[Root.m_nRight]);
						break;
					case opLE:
						return Compute(m_pBinaryTree[Root.m_nLeft])
							<= Compute(m_pBinaryTree[Root.m_nRight]);
						break;
					case opGT:
						return Compute(m_pBinaryTree[Root.m_nLeft])
							> Compute(m_pBinaryTree[Root.m_nRight]);
						break;
					case opGE:
						return Compute(m_pBinaryTree[Root.m_nLeft])
							>= Compute(m_pBinaryTree[Root.m_nRight]);
						break;
					case opEQ:
						return Compute(m_pBinaryTree[Root.m_nLeft])
							== Compute(m_pBinaryTree[Root.m_nRight]);
						break;
					case opNotEQ:
						return Compute(m_pBinaryTree[Root.m_nLeft])
							!= Compute(m_pBinaryTree[Root.m_nRight]);
						break;
					case opCOND:
						if ( Compute(m_pBinaryTree[Root.m_nLeft]) )
						{
							return Compute( m_pBinaryTree[m_pBinaryTree[Root.m_nRight].m_nLeft] );
						}
						else
						{
							return Compute( m_pBinaryTree[m_pBinaryTree[Root.m_nRight].m_nRight] );
						}
						break;
					default:
						return Root.m_nVal;
						break;
					}
				}

				long UpdateAndCmptRightVal()
				{
					UpdateVal();
					return Compute( m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nRight] );
				}

			} OverTowOpExpObj;

			memcpy( this, &OverTowOpExpObj, sizeof(void*) );
		}
		break;
	default:
		ASSERT(false);
		break;
	}
}

void CExp::UpdateVal( )
{
	ASSERT( m_spfunctionGetValByID!=NULL );

	for(long i=0;i< m_nVarNum;i++)\
	{\
		m_pBinaryTree[m_pszVar[i].m_nPos2Set].m_nVal = \
		m_spfunctionGetValByID( m_spCurrState, m_pszVar[i].m_nVarId);\
	}
}

void CExp::SetFuncGetValById( long (*pGetValById)(const ctaState *pState, int nValId) )
{
	m_spfunctionGetValByID = pGetValById;
}

long CExp::Compute()
{
	ASSERT( false );//ʹ
	return printf( "Exp Compute\n" );
}

long (*CExp::m_spfunctionGetValByID)(const ctaState *pState, int nValId) = NULL;

long CExp::UpdateValAndCompute()
{
	UpdateVal();
	return Compute();
}

#ifndef ACCINI_BUFFER_SIZE
#define ACCINI_BUFFER_SIZE	1024*16
#endif
#if!defined(FILE_END) 
#define FILE_END 2
#endif
#if!defined(FILE_BEGIN)
#define FILE_BEGIN 0
#endif
#if!defined(FILE_CURR)
#define FILE_CURR 1
#endif

bool CExp::OrientationReadPos(FILE *&fp, const char *lpAppName, const char *lpKeyName)
{
	char *pBuffer = new char[ACCINI_BUFFER_SIZE];
	char *pValidPos = NULL;//ոЧλ

	bool bFind = false;
	char *pTemp = NULL;

	while ( true )
	{
		fgets( pBuffer, ACCINI_BUFFER_SIZE, fp );

		if ( feof(fp) )
		{
			break;
		}

		pValidPos = pBuffer;
		
		while ( *pValidPos == ' '||*pValidPos == '\t' )
		{
			pValidPos++;
		}

		pTemp = pValidPos + strlen(pValidPos) - 1;

		while ( *pTemp == ' ' || *pTemp == '\t' || *pTemp == '\n' || *pTemp == '\r' )
		{
			*pTemp -- = '\0';
		}

		if ( pValidPos[0]!='[' || pValidPos[strlen(pValidPos)-1]!=']' )
		{
			continue;
		}

		pValidPos[strlen(pValidPos)-1] = '\0';

		if ( strcmp( lpAppName, pValidPos+1 )!=0 )
		{
			continue;
		}

		while ( true )
		{
			long nfPos = ftell( fp );
			
			fgets( pBuffer, ACCINI_BUFFER_SIZE, fp );

			pValidPos = pBuffer;
			
			while ( *pValidPos == ' '||*pValidPos == '\t' )
			{
				pValidPos++;
			}

			if ( pValidPos[0] == '[' )//Ѿҵһڵ
			{
				break;
			}

			if ( ( pTemp = strchr(pValidPos, '=')) == NULL )
			{
				if ( feof(fp) )
				{
					break;
				}

				continue;
			}

			*pTemp = '\0';
			
			if ( strcmp(pValidPos, lpKeyName) != 0 )
			{
				if ( feof(fp) )
				{
					break;
				}
				
				continue;
			}

			bFind = true;
			fseek( fp, nfPos+pTemp+1-pBuffer, FILE_BEGIN );
			break;
		}

		if ( bFind )
		{
			break;
		}
	}

	delete []pBuffer;

	return bFind;
}

#ifndef WIN32
char* ltoa(long i, char *s, int BASE )
{
	int m, n, sign;
	
	/* ȸisignԼi */ 
	if (i < 0) {
		sign = 0;
		i = -i;
	} else
		sign = 1;
	
	/* õBASEΪʱiλn */
	for (n = 1, m = i; m /= BASE; n++)
		;
	/* Ҫַ'\0' */
	if (sign)
		s[n--] = '\0';
	else        
		s[n + 1] = '\0';
	
	/* λBASEΪʱλϵ */
	do
	s[n--] = i % BASE + '0';
	while (i /= BASE);
	if (!sign)
		s[0] = '-';
	
	return s;
}
#endif

CExp::CExp(const CExp &Other)
{
	memcpy( this, &Other, sizeof(Other) );//ǳ
	m_pBinaryTree = new CExp::CNode[m_nTreeSize];//
	m_pszVar = new VAR[m_nVarNum];
	memcpy(m_pBinaryTree, Other.m_pBinaryTree, sizeof(CNode)*m_nTreeSize );
	memcpy(m_pszVar, Other.m_pszVar, sizeof(VAR)*m_nVarNum );
}

CExp &CExp::operator =(const CExp &Other)
{
	memcpy( this, &Other, sizeof(Other) );//ǳ
	m_pBinaryTree = new CExp::CNode[m_nTreeSize];//
	m_pszVar = new VAR[m_nVarNum];
	memcpy(m_pBinaryTree, Other.m_pBinaryTree, sizeof(CNode)*m_nTreeSize );
	memcpy(m_pszVar, Other.m_pszVar, sizeof(VAR)*m_nVarNum );
	return *this;
}

long CExp::GetLeftVarId()
{
	ASSERT( m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nLeft].m_OpType == opINVALIDOP );
	return m_pBinaryTree[m_pBinaryTree[m_nRoot].m_nLeft].m_nId;
}

OPTYPE CExp::GetRootOpType()
{
	return m_pBinaryTree[m_nRoot].m_OpType;
}

CExp CExp::GetRightSubExp()
{
	CExp Exp = *this;
	
	for(long i=0; i<Exp.m_nVarNum; i++)
	{
		if ( Exp.m_pszVar[i].m_nPos2Set == Exp.m_pBinaryTree[Exp.m_nRoot].m_nLeft )
		{
			Exp.m_pszVar[i] = Exp.m_pszVar[Exp.m_nVarNum-1];//ɾʽ߱
			Exp.m_nVarNum --;
			break;
		}
	}

	Exp.m_nRoot = Exp.m_pBinaryTree[m_nRoot].m_nRight;
	return Exp;
}

long CExp::UpdateAndCmptRightVal()
{
	ASSERT( FALSE );//ô˺
	return 0;
}

#else
#include "fpta.tab.h"
#include "decls.h"
#include "List.h"
#include "GetLU.h"
#include "MyList.h"

long CExp::AddLeftChild(long nRoot, long nVal)
{
	ASSERT( m_nStackTop < m_nDefaultSize );
	m_pBinaryTree[nRoot].m_nLeft = m_nStackTop;
	m_pBinaryTree[m_nStackTop].m_nVal = nVal;
	return m_nStackTop++;
}

long CExp::AddRightChild(long nRoot, long nVal)
{
	ASSERT( m_nStackTop < m_nDefaultSize );
	m_pBinaryTree[nRoot].m_nRight = m_nStackTop;
	m_pBinaryTree[m_nStackTop].m_nVal = nVal;
	return m_nStackTop++;
}

void CExp::Clear()
{
	for( long i=0; i<m_nStackTop; i++ )
	{
		m_pBinaryTree[i].Clear();
	}

	m_nStackTop = 0;
	m_nVarStackTop = 0;
}

void CExp::AddVar2Set(long nVarPos, long nVarId)
{
	ASSERT( m_nVarStackTop+1 < m_nDefaultVarSize );
	m_pszVar[m_nVarStackTop].m_nPos2Set = nVarPos;
	m_pszVar[m_nVarStackTop].m_nVarId = nVarId;
	m_pBinaryTree[nVarPos].m_nVal = nVarId;
	m_nVarStackTop++;
}

long CExp::AddNode(long nVal,OPTYPE OpType /* = INVALIDOP */)
{
	ASSERT( m_nStackTop+1 < m_nDefaultSize );
	m_pBinaryTree[m_nStackTop].m_nVal = nVal;
	m_pBinaryTree[m_nStackTop].m_OpType = OpType;
	return m_nStackTop++;
}

void CExp::SetLeftChild(long nRoot, long nLeftChild )
{
	m_pBinaryTree[nRoot].m_nLeft = nLeftChild;
}

void CExp::SetRightChild(long nRoot, long nRightChild )
{
	m_pBinaryTree[nRoot].m_nRight = nRightChild;
}

void CExp::writeINI(FILE *fp, const long nExpId )
{
	OrientationWritePos( fp ); 
	fprintf( fp, " exp-%d=",nExpId);
	fwrite( &m_nVarStackTop, sizeof(m_nVarStackTop), 1, fp );
	fwrite( &m_nStackTop, sizeof(m_nStackTop), 1, fp );
	fwrite( &m_nRoot, sizeof(m_nRoot), 1, fp );
	fwrite( m_pszVar, sizeof(VAR), m_nVarStackTop, fp );
	fwrite( m_pBinaryTree, sizeof(CNode), m_nStackTop, fp );
	fprintf( fp, "\r\n");
}

void CExp::AppointRoot(long nRoot)
{
	m_nRoot = nRoot;
}

void CExp::OrientationWritePos(FILE *&fp)
{
	
}

CExp g_Exp;

long AddLeftChild(long nRoot, long nVal){return g_Exp.AddLeftChild(nRoot, nVal);}
long AddRightChild(long nRoot, long nVal){return g_Exp.AddRightChild(nRoot,nVal);}
long AddNode(long nVal, OPTYPE chOp){return g_Exp.AddNode(nVal, chOp);}
void AddVar2Set(long nVarPos,long nVarId){g_Exp.AddVar2Set(nVarPos, nVarId);}
void expClear(){g_Exp.Clear();}
long g_nCurrRoot = 0;
void SetLeftChild(long nRoot, long nLeftChild ){g_Exp.SetLeftChild(nRoot,nLeftChild);}
void SetRightChild(long nRoot, long nRightChild ){g_Exp.SetRightChild(nRoot,nRightChild);}
void expWriteINI(long nExpId){g_Exp.writeINI( g_fpOutputObj,nExpId);}
long g_nTotalExpNum = 0;
void expSetExpNum(long nExpNum){ g_nTotalExpNum = nExpNum; }
long expGetExpNum(){ return g_nTotalExpNum; }
void AppointRoot( long nRoot ){g_Exp.AppointRoot(nRoot);}

OPTYPE translateCmpOp(const char *strOptype )
{
	if ( strcmp(strOptype, ">" ) == 0 || strcmp(strOptype, "&gt;" ) == 0 ) 
	{
		return opGT;
	}
	else
		if ( strcmp(strOptype, ">=" ) == 0 || strcmp(strOptype, "&gt;=" ) == 0 )  
		{
			return opGE;
		}
		else
			if ( strcmp(strOptype, "<" ) == 0 || strcmp(strOptype, "&lt;" ) == 0 )   
			{
				return opLT;
			}
			else
				if ( strcmp(strOptype, "<=" ) == 0 || strcmp(strOptype, "&lt;=" ) == 0 )   
				{
					return opLE;
				}
				else
					if ( strcmp(strOptype, "==" ) == 0 ) 
					{
						return opEQ;
					}
					else
						if ( strcmp(strOptype, "!=" ) == 0 ) 
						{
							return opNotEQ;
						}
						else
						{
							ASSERT( FALSE );
							return opINVALIDOP;
						}
}

void dealWithConst(void *pYYSTYPE_Exp)
{
	YYSTYPE *pExp = (YYSTYPE*)pYYSTYPE_Exp;

	if ( pExp->m_ExpType.m_bConst )
	{
		fflush( g_fpInput2Xml );
		fseek( g_fpInput2Xml, 0-pExp->m_ExpType.m_ni2xExpLen, FILE_CURR );
		pExp->m_ExpType.m_ni2xExpLen =
			fprintf( g_fpInput2Xml, "%d", pExp->m_ExpType.m_nVal );
	}
}

class CExpNode
{
public:
	std::string m_text;
	OPTYPE m_OpType;
	CExpNode *m_pLeft;
	CExpNode *m_pRight;

	CExpNode()
	{
		m_pLeft = m_pRight = NULL;
	}

protected:
private:
};

#ifndef max
#define max(a,b) ((a)>(b)?(a):(b))
#endif

CMyList< CExpNode > g_ExpNodeList;

const char *translateCmpOp( const OPTYPE opType, std::string &strStore )
{
	switch( opType )
	{
	case opSUB:
		strStore = "-";
		break;
	case opNOT:
		strStore = "!";
		break;
	case opADD:
		strStore = "+";
		break;
	case opMUL:
		strStore = "*";
		break;
	case opDIV:
		strStore = "/";
		break;
	case opMOD:
		strStore = "%";
		break;
	case opAND:
		strStore = "&amp;&amp;";
		break;
	case opOR:
		strStore = "||";
		break;
	case opLT:
		strStore = "&lt;";
		break;
	case opLE:
		strStore = "&lt;=";
		break;
	case opGT:
		strStore = "&gt;";
		break;
	case opGE:
		strStore = "&gt;=";
		break;
	case opEQ:
		strStore = "==";
		break;
	case opNotEQ:
		strStore = "!=";
		break;
	case opCOND1:
		strStore = "?";
		break;
	case opCOND2:
		strStore = ":";
		break;
	default:
		ASSERT(FALSE);
		break;
	}
	return strStore.data();
}

void *MakeExpNode(OPTYPE opType, void *pLeftOp, void *pRightOp, void *pParentOp )
{
	CExpNode *pExpNode = g_ExpNodeList.push_back();
	pExpNode->m_OpType = opType;
	YYSTYPE *pLeft = (YYSTYPE*)pLeftOp;
	YYSTYPE *pRight = (YYSTYPE*)pRightOp;
	YYSTYPE *pParent = (YYSTYPE*)pParentOp;

	if ( pParent->m_ExpType.m_bConst )
	{
		pExpNode->m_OpType = opINVALIDOP;
		std::ostringstream   s;   
		s   <<   pParent->m_ExpType.m_nVal;   
		pExpNode->m_text = s.str();
		return g_ExpNodeList.tail();
	}
	
	switch(opType) 
	{
	case opINVALIDOP:
		ASSERT(sizeof(void*)>=max(sizeof(char),sizeof(int)));//ȷvoid*벻жʧ
		switch((char)(int)pLeftOp)
		{
		case 'n'://
			{
				std::ostringstream   s;   
				s   <<   (int)pRightOp;   
				pExpNode->m_text = s.str();
			}
			break;
		case 'v'://
			pExpNode->m_text = (char*)pRightOp;
			break;
		default:
			ASSERT(FALSE);
			break;
		}
		break;
	
	case opSUB:
	case opNOT:
	case opADD:
	case opMUL:
	case opDIV:
	case opMOD:
	case opAND:
	case opOR:
	case opLT:
	case opLE:
	case opGT:
	case opGE:
	case opEQ:
	case opNotEQ:
	case opCOND1:
	case opCOND2:
		pExpNode->m_pLeft = (CExpNode*)pLeft->m_ExpType.m_pExpRoot;
		pExpNode->m_pRight = (CExpNode*)pRight->m_ExpType.m_pExpRoot;
		translateCmpOp( opType,	pExpNode->m_text );
		break;
	default:
		ASSERT(FALSE);
		break;
	}

	return g_ExpNodeList.tail();
}

void output2xml( void *pRoot )
{
	CExpNode *pExpRoot = (CExpNode*)pRoot;

	ASSERT( pExpRoot!=NULL  );

	switch( pExpRoot->m_OpType )
	{
	case opINVALIDOP:
		i2x( "%s", pExpRoot->m_text.data() );
		break;
	case opNOT:
		ASSERT(pExpRoot->m_pLeft == NULL );
		i2x( "%s", pExpRoot->m_text.data() );
		i2x("(");
		output2xml( pExpRoot->m_pRight );
		i2x(")");
		break;
	case opSUB:
		if ( pExpRoot->m_pLeft == NULL )
		{
			i2x( "%s", pExpRoot->m_text.data() );
			i2x("(");
			output2xml( pExpRoot->m_pRight );
			i2x(")");
			break;
		}
	case opADD:
	case opMUL:
	case opDIV:
	case opMOD:
	case opAND:
	case opOR:
	case opLT:
	case opLE:
	case opGT:
	case opGE:
	case opEQ:
	case opNotEQ:
		i2x("(");
		output2xml( pExpRoot->m_pLeft );
		i2x(")");
		i2x( "%s", pExpRoot->m_text.data() );
		i2x("(");
		output2xml( pExpRoot->m_pRight );
		i2x(")");
		
		break;
	case opCOND1:
		i2x("(");
		output2xml( pExpRoot->m_pLeft );
		i2x(")");
		i2x( "%s", pExpRoot->m_text.data() );
		output2xml( pExpRoot->m_pRight );
		break;
	case opCOND2:
		i2x("(");
		output2xml( pExpRoot->m_pLeft );
		i2x(")");
		i2x( "%s", pExpRoot->m_text.data() );
		i2x("(");
		output2xml( pExpRoot->m_pRight );
		i2x(")");
		break;
	default:
		ASSERT(FALSE);
		break;
	}
}

void OutputAndClear1ExpTree(void *pRoot)
{
	ASSERT( pRoot!=NULL );
	YYSTYPE *pExpRoot = (YYSTYPE*)pRoot;
	ASSERT( (void*)pExpRoot == (void*)&pExpRoot->m_ExpType );//ȷCȡַȷ
	fseek( g_fpInput2Xml, 0-pExpRoot->m_ExpType.m_ni2xExpLen, FILE_CURR );
	output2xml( pExpRoot->m_ExpType.m_pExpRoot );
	g_ExpNodeList.Clear();
}

bool i2xExpHasClockVar(CExpNode *pExpRoot );

void CheckCmpExp(void *pRoot )
{
	ASSERT(pRoot!=NULL);
	CExpNode *pExpRoot = (CExpNode*)pRoot;

	switch( pExpRoot->m_OpType )
	{
	case opLT:
	case opLE:
	case opGT:
	case opGE:
	case opEQ:
	case opNotEQ:
		break;
	case opINVALIDOP:
	case opSUB:
	case opNOT:
	case opADD:
	case opMUL:
	case opDIV:
	case opMOD:
	case opAND:
	case opOR:
	case opCOND1:
	case opCOND2:
		err_num++;
		sprintf(temp_err[err_num-1],"ʽֱֻȽϱʽ");
		fprintf(stderr, "\r\n%s\r\n", temp_err[err_num-1]);
		input_error(err_num,"exp error!",lineno,0,temp_err[err_num-1]);
		break;
	default:
		ASSERT(FALSE);
		break;
	}

	if ( pExpRoot->m_pLeft == NULL )
	{
		ASSERT(pExpRoot->m_OpType==opINVALIDOP
			||pExpRoot->m_OpType==opSUB
			||pExpRoot->m_OpType==opNOT
			||pExpRoot->m_OpType==opCOND1
			||pExpRoot->m_OpType==opCOND2);
		return;
	}

	if ( pExpRoot->m_pLeft->m_OpType!=opINVALIDOP )
	{
		err_num++;
		sprintf(temp_err[err_num-1],"Ƚϱʽ߱ǵ");
		fprintf(stderr, "\r\n%s\r\n", temp_err[err_num-1]);
		input_error(err_num,"exp error!",lineno,0,temp_err[err_num-1]);
	}
	else
	{
		char chTemp = pExpRoot->m_pLeft->m_text[0];
		
		//ﲻжconst idΪconst idѾ
		if ( '0'<=chTemp && chTemp <= '9' )//Ǳid
		{
			err_num++;
			sprintf(temp_err[err_num-1],"Ƚϱʽ߱ǵǳ");
			fprintf(stderr, "\r\n%s\r\n", temp_err[err_num-1]);
			input_error(err_num,"exp error!",lineno,0,temp_err[err_num-1]);
		}
	}

	if ( i2xExpHasClockVar(pExpRoot->m_pRight) )
	{
		err_num++;
		sprintf(temp_err[err_num-1],"ʱӱڱʽұ");
		fprintf(stderr, "\r\n%s\r\n", temp_err[err_num-1]);
		input_error(err_num,"exp error!",lineno,0,temp_err[err_num-1]);
	}
}

bool i2xExpHasClockVar(CExpNode *pExpRoot )
{
	if ( pExpRoot == NULL )
	{
		return false;
	}

	return i2xIsClock( GetCurrProcessName(), pExpRoot->m_text.data() )
		|| i2xExpHasClockVar( pExpRoot->m_pLeft )
		|| i2xExpHasClockVar( pExpRoot->m_pRight );
}

#endif

void CExp::CNode::Clear()
{
	m_nLeft = m_nRight = -1;
	m_OpType = opINVALIDOP;//Ч
}

#ifdef _DEBUG
void CExp::dbgOutput()
{
	fprintf( stdout, "\n");
	dbgOutput( m_pBinaryTree[m_nRoot] );
}

void CExp::dbgOutput(CNode &Node )
{
#ifdef __EXP_USED_BY_COMPILER
#define VarNum m_nVarStackTop
#else
#define VarNum m_nVarNum
#endif
	if ( Node.m_OpType == opINVALIDOP )
	{
		for( long i=0; i<VarNum; i++ )
		{
			if ( &m_pBinaryTree[m_pszVar[i].m_nPos2Set] == &Node )
			{
				fprintf( stdout, "id%d", m_pszVar[i].m_nVarId );
				ASSERT( m_pBinaryTree[m_pszVar[i].m_nPos2Set].m_nId == m_pszVar[i].m_nVarId );
				return;
			}
		}

		fprintf( stdout, "%d", Node.m_nVal );
		return;
	}
	
	fprintf( stdout, "(" );

	if ( Node.m_OpType == opCOND )
	{
		dbgOutput( m_pBinaryTree[Node.m_nLeft] );
		fprintf( stdout, "?" );
		dbgOutput( m_pBinaryTree[m_pBinaryTree[Node.m_nRight].m_nLeft] );
		fprintf( stdout, ":" );
		dbgOutput( m_pBinaryTree[m_pBinaryTree[Node.m_nRight].m_nRight] );
	}
	else
	if ( Node.m_nLeft>=0 )
	{
		dbgOutput( m_pBinaryTree[Node.m_nLeft] );
	}
	
	switch( Node.m_OpType )
	{
	case opADD:
		fprintf( stdout, "+" );
		break;
	case opSUB:
		fprintf( stdout, "-" );
		break;
	case opMUL:
		fprintf( stdout, "*" );
		break;
	case opDIV:
		fprintf( stdout, "/" );
		break;
	case opMOD:
		fprintf( stdout, "%" );
		break;
	case opNOT:
		fprintf( stdout, "!" );
		break;
	case opAND:
		fprintf( stdout, "&&" );
		break;
	case opOR:
		fprintf( stdout, "||" );
		break;
	case opLT:
		fprintf( stdout, "<" );
		break;
	case opLE:
		fprintf( stdout, "<=" );
		break;
	case opGT:
		fprintf( stdout, ">" );
		break;
	case opGE:
		fprintf( stdout, ">=" );
		break;
	case opEQ:
		fprintf( stdout, "==" );
		break;
	case opNotEQ:
		fprintf( stdout, "!=" );
		break;
	case opCOND:
		break;
	default:
		for( long i=0; i<VarNum; i++ )
		{
			if ( &m_pBinaryTree[m_pszVar[i].m_nPos2Set] == &Node )
			{
				fprintf( stdout, "id%", m_pszVar[i].m_nVarId );
				break;
			}
		}

		fprintf( stdout, "%d", Node.m_nVal );
		break;
	}

	if ( Node.m_OpType == opCOND )
	{
	}
	else
	if ( Node.m_nRight>=0 )
	{
		dbgOutput( m_pBinaryTree[Node.m_nRight] );
	}
	
	fprintf( stdout, ")" );

	/*if ( Node.m_OpType == opSUB || Node.m_OpType == opUNIMUS
		||Node.m_OpType == opNOT )
	{
		if ( Node.m_nLeft>=0 )
		{
			dbgOutput( m_pBinaryTree[Node.m_nLeft] );
		}
		
		dbgOutput( m_pBinaryTree[Node.m_nRight] );
	}
	if ( Node.m_OpType != opINVALIDOP && Node.m_OpType!=opNOT
		&&Node.m_OpType != opUNIMUS && Node.m_OpType != opSUB
		&&Node.m_OpType != opCOND )
	{
		dbgOutput( m_pBinaryTree[Node.m_nLeft] );
		dbgOutput( m_pBinaryTree[Node.m_nRight] );
	}*/
}
#endif

long i2xComputeExp(const long nVal1, const long nVal2, const OPTYPE chOpType)
{
	switch( chOpType )
	{
	case '+':
		return nVal1+nVal2;
		break;
	case '-':
		return nVal1 -nVal2;
		break;
	case '*':
		return nVal1 *nVal2;
		break;
	case '/':
		if ( nVal2==0 )
		{
			fprintf( stderr, "0");
			return -1;
		}
		return nVal1 /nVal2;
		break;
	case '%':
		if ( nVal2==0 )
		{
			fprintf( stderr, "ܶ0ȡ");
			return -1;
		}
		return nVal1 %nVal2;
		break;
	case '&':
		return nVal1 && nVal2;
		break;
	case '|':
		return nVal1 || nVal2;
		break;
	default:
		ASSERT(FALSE);
		return -1;
		break;
	}
}

